package de.lmu.ifi.mfchords.utils;

import java.util.ArrayList;



import android.graphics.PointF;
import android.util.Log;

/**
 * This class keeps track of the hands approximate position and orientation over the tablet's surface
 * 
 */



public class HandModel {
	
	private Finger thumb = new Finger(Constants.THUMB, new PointF(0,0));
	private Finger index = new Finger(Constants.INDEX, new PointF(0,0));
	private Finger middle = new Finger(Constants.MIDDLE, new PointF(0,0));
	private Finger ring = new Finger(Constants.RING, new PointF(0,0));
	private Finger pinky = new Finger(Constants.PINKY, new PointF(0,0));
	
	
	private ArrayList<Finger> allFingers = new ArrayList<Finger>();
	private String chordBasis = Constants.ChordBasisNone; 
	private ArrayList<PointF> allPoints = new ArrayList<PointF>();
	
	private String recognizedHandPosture;
	
	HandModel(){
		allFingers.add(thumb);
		allFingers.add(index);
		allFingers.add(middle);
		allFingers.add(ring);
		allFingers.add(pinky);
	}
	
	
	private void resetPredictions(){
		for(Finger f: allFingers){
			f.reset();
		}
	}
	
	protected void reset(){
		this.chordBasis = Constants.ChordBasisNone;
		resetPredictions();
		this.recognizedHandPosture = Constants.Posture_None;
		
	}
	
	/**
	   * this method returns the relative measurements to the three points basis 1, basis 2 and point p.
	   * 
	   * 
	   * @param b1 touch position of first basis touch
	   * @param b2 touch position of second basis touch
	   * @param p touch position of the third touch   
	   * 
	   * @return RelativeMeasurement extracted relative measurements from the touch events.
	   */
	protected RelativeMeasurement GetRelativeMeasurements(PointF b1, PointF b2, PointF p) {
		
		allPoints.clear();
		allPoints.add(b1);
		allPoints.add(b2);
		allPoints.add(p);
		
		
		double maxValue = 4.8;
		
		
		RelativeMeasurement rm = new RelativeMeasurement();
		
		//RD values
		double D1 = Utils.distanceBetweenPoints(b1.x, b1.y, p.x, p.y);
		rm.setD1(D1);
		double D2 = Utils.distanceBetweenPoints(p.x, p.y, b2.x, b2.y);
		rm.setD2(D2);
		double D3 = Utils.distanceBetweenPoints(b1.x, b1.y, b2.x, b2.y);
		rm.setD3(D3);
		
		
		double RD1 = D1/D3;
		rm.setRD1(RD1/maxValue);
		double RD2 = D1/D2;
		rm.setRD2(RD2/maxValue);
		double RD3 = D2/D3;
		rm.setRD3(RD3/maxValue);
		
		
		double[] p0 = {p.x, p.y,0};
		double[] p1 = {b1.x, b1.y,0};
		double[] p2 = {b2.x, b2.y,0};

		
		double RA1 = Math.toDegrees(Utils.computeAngle(p1, p0, p2));
		rm.setRA1(RA1/360);
		double RA2 = Math.toDegrees(Utils.computeAngle(p0, p1, p2));
		rm.setRA2(RA2/360);
		double RA3 = Math.toDegrees(Utils.computeAngle(p2, p0, p1));
		rm.setRA3(RA3/360);
	
		
		double RPD1 = Utils.distPtLine(p, b2, b1) / Utils.distanceBetweenPoints(p.x, p.y, b2.x, b2.y);
		rm.setRPD1(RPD1/maxValue);
		double RPD2 = Utils.distPtLine(b1, b2, p)/Utils.distanceBetweenPoints(b1.x, b1.y, b2.x, b2.y);
		rm.setRPD2(RPD2/maxValue);
		double RPD3 = Utils.distPtLine(b1, p, b2)/Utils.distanceBetweenPoints(b1.x, b1.y, p.x, p.y);;
		rm.setRPD3(RPD3/maxValue);
	
		
	
		double RP1 = Utils.relativePosition(p, b2, b1); 
		rm.setRP1(RP1);
		
		double RP2 = Utils.relativePosition(b1, b2, p);
		rm.setRP2(RP2);
		
		double RP3 = Utils.relativePosition(b1, p, b2);
		rm.setRP3(RP3);
		
		
		return rm;
	}
										

	
	protected void predictFingerPositions(){
		if(this.chordBasis.equals(Constants.ChordBasis_TI)){
			this.middle.setPredictedPosition(calculateTIPositions(Constants.TI_RA_middle, Constants.TI_RD_middle));
			this.ring.setPredictedPosition(calculateTIPositions(Constants.TI_RA_ring, Constants.TI_RD_ring));
			this.pinky.setPredictedPosition(calculateTIPositions(Constants.TI_RA_pinky, Constants.TI_RD_pinky));
		}else if(this.chordBasis.equals(Constants.ChordBasis_TP)){
			this.index.setPredictedPosition(calculateTPPositions(Constants.TP_RA_index, Constants.TP_RD_index));
			this.middle.setPredictedPosition(calculateTPPositions(Constants.TP_RA_middle, Constants.TP_RD_middle));
			this.ring.setPredictedPosition(calculateTPPositions(Constants.TP_RA_ring, Constants.TP_RD_ring));
		}else{
			if(thumb.isPositionSet()){
				//3nfindex
				//calculate position of ring finger by information about the 3nfmiddle 
				//RA3 135.14
				//RD2 1.16
				//Log.v("myDebugTag","thumb not null ");
				
				PointF rotP = Utils.rotatePoint(middle.position, index.position, Constants.NF_RA_middle);
				Double[] m1 = {(double) (rotP.x - middle.position.x), (double) (rotP.y - middle.position.y)};
				PointF ringPos = new PointF(Double.valueOf(middle.position.x + Constants.NF_RD_middle * m1[0]).floatValue(), Double.valueOf(middle.position.y + Constants.NF_RD_middle * m1[1]).floatValue());
				this.ring.setPredictedPosition(ringPos);
				
				PointF rotPinkyP = Utils.rotatePoint(ring.getPredictedPosition(), middle.position, Constants.NF_RA_ring);
				Double[] mPinky = {(double) (rotPinkyP.x - ring.getPredictedPosition().x), (double) (rotPinkyP.y - ring.getPredictedPosition().y)};
				PointF pinkyPos = new PointF(Double.valueOf(ring.getPredictedPosition().x + Constants.NF_RD_ring * mPinky[0]).floatValue(), Double.valueOf(ring.getPredictedPosition().y + Constants.NF_RD_ring * mPinky[1]).floatValue());
				this.pinky.setPredictedPosition(pinkyPos);
				
				//the same for pinky
				//this.pinky = new Finger("pinky", calculatePosition(ring, middle, TNF_relA_tnfring, TNF_relD_tnfring));
				
				
			}else if(pinky.isPositionSet()){
				//3nfring RA1_middle 20.72
				//RD3 = 0.59
				
				//rotate middle around ring finger
				PointF rotMiddle = Utils.rotatePoint(ring.position, middle.position, Constants.NF_RA_ring_index);
				Double[] mRI = {(double) (rotMiddle.x - ring.position.x), (double) (rotMiddle.y - ring.position.y)};
				PointF indexPos = new PointF(Double.valueOf(ring.position.x + Constants.NF_RD_ring_index * mRI[0]).floatValue(), Double.valueOf(ring.position.y + Constants.NF_RD_ring_index * mRI[1]).floatValue());
				this.index.setPredictedPosition(indexPos);
				
				//rotate index around middle finger
				PointF rotIndex = Utils.rotatePoint(middle.position, index.getPredictedPosition(), Constants.NF_RA_ring_thumb);
				Double[] mMT = {(double) (rotIndex.x - middle.position.x), (double) (rotIndex.y - middle.position.y)};
				PointF thumbPos = new PointF(Double.valueOf(middle.position.x + Constants.NF_RD_ring_thumb * mMT[0]).floatValue(), Double.valueOf(middle.position.y + Constants.NF_RD_ring_thumb * mMT[1]).floatValue());
				this.thumb.setPredictedPosition(thumbPos);
				
				//this.index = new Finger("index", calculatePosition(ring, middle, 24, 0.74f));
				//this.thumb = new Finger("index", calculatePosition(middle, index, 140, 0.43f));
				//PointF rotMiddle = Utils.rotatePoint(ring.position, middle.position, 20.72);
				//Double[] mm1 = {(double) (rotPinky.x - ring.position.x), (double) (rotPinky.y - ring.position.y)};
				
				//PointF pinkyPos = new PointF(Double.valueOf(ring.position.x + TNF_relD_tnfring * m1[0]).floatValue(), Double.valueOf(ring.position.y + TNF_relD_tnfring * m1[1]).floatValue());
				//this.pinky = new Finger("pinky", pinkyPos);
				
			}else {
				//3nf middle
				
				//rotate index around middle finger
				PointF rotIndex = Utils.rotatePoint(middle.position, index.position, Constants.NF_RA_ring_thumb);
				Double[] mMT = {(double) (rotIndex.x - middle.position.x), (double) (rotIndex.y - middle.position.y)};
				PointF thumbPos = new PointF(Double.valueOf(middle.position.x + Constants.NF_RD_ring_thumb * mMT[0]).floatValue(), Double.valueOf(middle.position.y + Constants.NF_RD_ring_thumb * mMT[1]).floatValue());
				//this.thumb = new Finger("thumb", thumbPos);
				this.thumb.setPredictedPosition(thumbPos);
				
				
				PointF rotPinkyP = Utils.rotatePoint(ring.position, middle.position, Constants.NF_RA_ring);
				Double[] mPinky = {(double) (rotPinkyP.x - ring.position.x), (double) (rotPinkyP.y - ring.position.y)};
				PointF pinkyPos = new PointF(Double.valueOf(ring.position.x + Constants.NF_RD_ring * mPinky[0]).floatValue(), Double.valueOf(ring.position.y + Constants.NF_RD_ring * mPinky[1]).floatValue());
				//this.pinky = new Finger("pinky", pinkyPos);
				this.pinky.setPredictedPosition(pinkyPos);
			}

		}
			
	}
	
	
	
	
	protected void updateRemainingFingerPositions(String mfc){
		this.resetPredictions();
		
		if(mfc.startsWith("TI")){
			if(allPoints.size() > 2){
				this.thumb.setPosition(allPoints.get(1));
				this.index.setPosition(allPoints.get(0));
				if(mfc.endsWith(Constants.MIDDLE)){
					this.middle.setPosition(allPoints.get(2));
				}else if(mfc.endsWith(Constants.RING)){
					this.ring.setPosition(allPoints.get(2));
				}else if(mfc.endsWith(Constants.PINKY)){
					this.pinky.setPosition(allPoints.get(2));
				}
			}
		}else if(mfc.startsWith("TP")){
			if(allPoints.size() > 2){
				this.thumb.setPosition(allPoints.get(0));
				this.pinky.setPosition(allPoints.get(1));
				if(mfc.endsWith(Constants.INDEX)){
					this.index.setPosition(allPoints.get(2));
				}else if(mfc.endsWith(Constants.MIDDLE)){
					this.middle.setPosition(allPoints.get(2));
				}else if(mfc.endsWith(Constants.RING)){
					this.ring.setPosition(allPoints.get(2));
				}
			}
		}else if(mfc.startsWith("NF")){
			if(allPoints.size() > 2){
				if(mfc.endsWith(Constants.INDEX)){
					this.thumb.setPosition(allPoints.get(0));
					this.middle.setPosition(allPoints.get(1));
					this.index.setPosition(allPoints.get(2));
				}else if(mfc.endsWith(Constants.MIDDLE)){
					this.index.setPosition(allPoints.get(0));
					this.ring.setPosition(allPoints.get(1));
					this.middle.setPosition(allPoints.get(2));
				}else if(mfc.endsWith(Constants.RING)){
					this.middle.setPosition(allPoints.get(0));
					this.pinky.setPosition(allPoints.get(1));
					this.ring.setPosition(allPoints.get(2));
				}
			}
		}
		
		predictFingerPositions();
		
		}
		
	protected PointF calculateTPPositions(float tp_relA_finger, float tp_relD_finger){
		
		
		Double[] v = new Double[]{Double.valueOf(pinky.position.x - thumb.position.x), Double.valueOf(pinky.position.y - thumb.position.y)};
		
		
		//position relative between thumb and pinky
		PointF rel_pos_TP = new PointF(Double.valueOf(thumb.position.x + tp_relD_finger * v[0]).floatValue(), Double.valueOf(thumb.position.y + tp_relD_finger * v[1]).floatValue());
		
		//rotate around thumb
		PointF rotP = Utils.rotatePoint(thumb.position, rel_pos_TP, tp_relA_finger);
    	
    	return rotP;
    	
	}
	
	protected PointF calculateTIPositions(float ti_relA_finger, float ti_relD_finger){
		PointF rotP = Utils.rotatePoint(index.position, thumb.position, ti_relA_finger);
		Double[] m1 = {(double) (rotP.x - index.position.x), (double) (rotP.y - index.position.y)};
		
		PointF middlePos = new PointF(Double.valueOf(index.position.x + ti_relD_finger * m1[0]).floatValue(), Double.valueOf(index.position.y + ti_relD_finger * m1[1]).floatValue()); 
		return middlePos;
		
	}
	
	protected void setTIBasis(ArrayList<PointF> b){
		this.thumb.setPosition(b.get(0));
		this.index.setPosition(b.get(1));
	}
	
	protected void setTPBasis(ArrayList<PointF> b){
		this.thumb.setPosition(b.get(0));
		this.pinky.setPosition(b.get(1));
	}
	
	protected void setBasis(ArrayList<PointF> b){
		if(this.chordBasis.equals(Constants.ChordBasis_TI)){
			this.setTIBasis(b);
		}else if(this.chordBasis.equals(Constants.ChordBasis_TP)){
			this.setTPBasis(b);
		}
	}


	public String getRecognizedHandPosture() {
		return recognizedHandPosture;
	}


	protected void setRecognizedHandPosture(String recognizedHandPosture) {
		this.recognizedHandPosture = recognizedHandPosture;
	}
	
	public ArrayList<Finger> getAllFingers(){
		return allFingers;
	}
	
	public String getChordBasis(){
		return chordBasis;
	}
	
	public void setChordBasis(String cb){
		this.chordBasis = cb;
	}
	
	
	}